#ifndef DECAYHEAP_HPP_
#define DECAYHEAP_HPP_

#include<vector>
#include<cassert>
#include "Assertion.hpp"

#define HEAP_INVALID_INDEX -1




/**
 * Heap structure that ranks all elements according to some activity number.
 * It allows for incrementation of the activity number of single elements and moreover, all activities can be decremented at once.
 * KeyType  - has to implement method 'int idx()' and has to be constructable by 'KeyType(int index)'
 * ValueType - type to rank elements has to be comparable by operators > and >= value has to be increasable by operator += */
template<class KeyType, class ValueType >
class DecayHeap
{
    /**Maximal rank value*/
    const ValueType max;
    const ValueType min;


    ValueType       increment;

    const ValueType decay;


    /**Internal data-structure to store the information for one element.*/
    struct ElementInfo{

        /**index of this element in the heap*/
        int index_in_heap;

        /**the rank value of an element*/
        ValueType value;

        /**Constructs an invalid emement.*/
        inline ElementInfo()                        :index_in_heap(HEAP_INVALID_INDEX),value(0){};

        /**Constructs an elemnt with the specified index with the given rank value*/
        inline ElementInfo(int index, ValueType v)  :index_in_heap(index),value(v){};

        /**No copying allowed.*/
        inline ElementInfo(const ElementInfo& e)    :index_in_heap(e.index_in_heap),value(e.value){};

        /**No assignment allowed.*/
        const ElementInfo& operator=(const ElementInfo& e){assert(false); return (ElementInfo&)e;};
    };


/*###  Class Functions ###*/
private:
    /**Return the index of the parent of an element.*/
    static inline int   parent        (int index)     {return (index-1) >> 1;};

    /**Return the index of the left child of an element.*/
    static inline int   leftChild     (int index)     {return (index << 1) + 1;};

    /**Return the index of the right child of an element.*/
    static inline int   rightChild    (int index)     {return (index << 1) + 2;};


/*###  Member Variables ###*/
private:
    /**Holds an ElementInfo for each element in the heap.*/
    std::vector<ElementInfo> element_info;

    /**The actual heap. Each element of the heap vector is an index of the original element.*/
    std::vector<int> heap;



/*###  Member Functions ###*/
private:
    /**No copying.*/
    inline              DecayHeap(const DecayHeap& h);

    /**No assignment.*/
    inline const        DecayHeap<KeyType,ValueType>& operator=(const DecayHeap<KeyType,ValueType>& h);

    /**Auxiliary function to check if the first element has a higher rank than the second one.*/
    inline bool         greater(int el_a, int el_b)
                        {return element_info[el_a].value > element_info[el_b].value;};

    /**Auxiliary function to check if the first element has a higher or equal rank as the second one.*/
    inline bool         greaterEq(int el_a, int el_b)
                        {return element_info[el_a].value >= element_info[el_b].value;};

    /**Moves an element up the heap until the heap property is valid.
     * @param element_id: indicates the id of the element to move
     * @param heap_index: indicates the current index in the heap
     * @returnt the new pos of the given element*/
    inline int          moveUp(const int element_id, int heap_index);

    /**Moves an element down the heap until the heap property is valid.
     * @param element_id: indicates the id of the element to move
     * @param heap_index: indicates the current index in the heap*/
    inline void         moveDown(const int element_id, int heap_index);

    /**Returns element id of the element that was at the specified position pos.
     * Does not ensure the heap property*/
    inline int          lazyRemoveAtPos(uint pos);


    inline bool         checkHeapProperty(uint heap_index);


public:

    /**Constructor
     * @param MAX: the maximal rank value of an element*/
    inline              DecayHeap(ValueType MAX,ValueType MIN, ValueType dec):
                        max(MAX),min(MIN),increment(MIN),decay(dec){ };

    /**Constructor
     * @param MAX: the maximal rank value of an element
     * @param capacity: the amount of space to reserve */
    inline              DecayHeap(ValueType MAX,ValueType MIN,ValueType dec, int capacity):
						max(MAX),min(MIN),increment(MIN),decay(dec),element_info(capacity),heap(capacity){ };

    /**Destructor*/
    inline              ~DecayHeap() { };

    /**Reserve enough space for the specified number of elements*/
    inline void         reserve(uint capacity);

    inline KeyType      removeAtPos(uint i);

    /**Get the element with the highest rank value and remove it from the heap.*/
    inline KeyType      removeMax();

    /**Increase the rank value for the specified element.
     * This is even possible if the element is currently not in the heap but has been removed before.*/
    inline bool         bump(const KeyType& element);

    /**Checks if the specified element is contained in the heap.*/
    inline bool         contains(const KeyType& element);

    /**Insert a new element in the heap.
     * @param element: a reference to the element to be inserted
     * @param initial_value: the initial rank value that is 0 per default.*/
    inline void         addNewElement(const KeyType& element, ValueType initial_value = 0);

    /**Insert a new element in the heap. This function won't arrange the heap property itself.
    * It has to be ensured by the user that the heap property holds, for instance if the initial_value is not greater than the one of all
    * previously added elements.
    * @param element: a reference to the element to be inserted
    * @param initial_value: the initial rank value that is 0 per default.*/
    inline int          lazyAddNewElement(const KeyType& element, ValueType initial_value = 0);

    /**Reactivates an element that was previously removed by the function removeMax or removeAtPos.*/
    inline void         restoreElement(const KeyType& element);

    inline void 		remove(const KeyType& element);

    inline void         decayAll()              {increment *= decay; };
    /**Return the size of the heap*/
    inline uint32_t     size()                  {return heap.size();};

    /**Return the current rank value of the specified element*/
    inline ValueType    getValue(const KeyType& element)
                        {assert(element.idx() < element_info.size());return element_info[element.idx()].value;}

    /**Checks if the heap is empty*/
    inline bool         empty()                 {return heap.empty();}

    inline KeyType      operator[](uint i)      {assert(i < heap.size()); return KeyType(heap[i]);}
};





//Inline Implementation


template<class KeyType, class ValueType>
int DecayHeap<KeyType, ValueType>::moveUp(const int element_id, int heap_index){
    assert(heap_index < (int)heap.size() && heap[heap_index] == element_id);
    int parent_index;
    while(  heap_index != 0
        && (parent_index = parent(heap_index), greater(element_id,heap[parent_index]))){

        heap[heap_index]                                =   heap[parent_index];
        element_info[heap[heap_index]].index_in_heap    =   heap_index;
        heap_index                                      =   parent_index;
    }
    heap[heap_index]                                    =   element_id;
    element_info[element_id].index_in_heap              =   heap_index;
    return heap_index;
};

template<class KeyType, class ValueType>
void DecayHeap<KeyType, ValueType>::moveDown(const int element_id, int heap_index){
    assert(heap_index < (int)heap.size() && heap[heap_index] == element_id);
    unsigned int child_index;
    while((child_index = leftChild(heap_index), child_index < heap.size()))
    {
        //get the child with the higher rank
        if(child_index+1 < heap.size() && greater(heap[child_index+1],heap[child_index]))
            ++child_index; //the right one is bigger

        if(greaterEq(element_id,heap[child_index]))
            break;

        heap[heap_index]                                =   heap[child_index];
        element_info[heap[heap_index]].index_in_heap    =   heap_index;
        heap_index                                      =   child_index;
    }
    heap[heap_index]                                    =   element_id;
    element_info[element_id].index_in_heap              =   heap_index;
};


template<class KeyType, class ValueType>
void DecayHeap<KeyType, ValueType>::reserve(uint capacity){
    if(capacity >= element_info.size()){
        element_info.resize(capacity);
        heap.reserve(capacity);}
}


//has to be ensured that not empty
template<class KeyType, class ValueType>
int DecayHeap<KeyType, ValueType>::lazyRemoveAtPos(uint pos){
    assert(pos < heap.size());

    int e_idx                               = heap[pos];
    int last                                = heap.back();
    heap[pos]                               = last;
    element_info[last].index_in_heap        = pos;
    element_info[e_idx].index_in_heap       = HEAP_INVALID_INDEX; //we leave value unchanged for readding
    heap.pop_back();

    return e_idx;
}

template<class KeyType, class ValueType>
KeyType DecayHeap<KeyType, ValueType>::removeAtPos(uint pos){
    int ret_idx = lazyRemoveAtPos(pos);
    if(heap.size() > 1 && pos < heap.size()){
        pos = moveUp(heap[pos],pos);
        moveDown(heap[pos],pos);}
    sanity(checkHeapProperty(0));
    return KeyType(ret_idx);
}


template<class KeyType, class ValueType>
KeyType DecayHeap<KeyType, ValueType>::removeMax(){
    int ret_idx = lazyRemoveAtPos(0);
    if(heap.size() > 1)  moveDown(heap[0],0);

    sanity(checkHeapProperty(0));
    return KeyType(ret_idx);
}


template<class KeyType, class ValueType>
bool DecayHeap<KeyType, ValueType>::bump(const KeyType& element){
    //assert(contains(element));

    ElementInfo& e_info = element_info[element.idx()];
    e_info.value += increment;

    if(e_info.index_in_heap != HEAP_INVALID_INDEX)
        moveUp(element.idx(),e_info.index_in_heap);

    if(e_info.value > max){
    	const typename ValueType::denominator_t w = min.getLogOfDenominator();
        increment.weaken(w);
        int i = element_info.size();
        while(i--) element_info[i].value.weaken(w);
        return false;
    }
    return true;
}



template<class KeyType, class ValueType>
bool DecayHeap<KeyType, ValueType>::contains(const KeyType& element){
    assert( element.idx() < element_info.size());
    return element_info[element.idx()].index_in_heap != HEAP_INVALID_INDEX;
}

template<class KeyType, class ValueType>
int DecayHeap<KeyType, ValueType>::lazyAddNewElement(const KeyType& element, ValueType initial_value){

    assert(element_info.size() > element.idx());
    assert(element_info[element.idx()].index_in_heap == HEAP_INVALID_INDEX);

    ElementInfo& e_info = element_info[element.idx()];
    e_info.value = initial_value;
    e_info.index_in_heap = heap.size();
    heap.push_back(element.idx());

    return e_info.index_in_heap;
};

template<class KeyType, class ValueType>
void DecayHeap<KeyType, ValueType>::addNewElement(const KeyType& element, ValueType initial_value){

    if(element.idx() >= (int)element_info.size())
        element_info.resize(element.idx() + 1);

    moveUp(element.idx(),lazyAddNewElement(element,initial_value));

};

template<class KeyType, class ValueType>
void DecayHeap<KeyType, ValueType>::restoreElement(const KeyType& element){
    assert(element_info.size() > element.idx());

    ElementInfo& e_info = element_info[element.idx()];
    e_info.index_in_heap = heap.size();
    heap.push_back(element.idx());

    moveUp(element.idx(),e_info.index_in_heap);
};



template<class KeyType, class ValueType>
void DecayHeap<KeyType, ValueType>::remove(const KeyType& element){
    assert(element.idx() >= 0 && element.idx() < (int)element_info.size());
    assert(element_info[element.idx()].index_in_heap != HEAP_INVALID_INDEX);
    removeAtPos(element_info[element.idx()].index_in_heap);
}

template<class KeyType, class ValueType>
bool DecayHeap<KeyType, ValueType>::checkHeapProperty(uint i){
    return i >= heap.size()
          || ((i == 0 || greaterEq(heap[parent(i)],heap[i])) && checkHeapProperty(leftChild(i)) && checkHeapProperty(rightChild(i)));
}

#define decay_heap_hpp_version "DecayHeap.hpp:      $Revision: 1.1 $"


#endif /*DECAYHEAP_HPP_*/


